#ifndef		NNODE_H
#define		NNODE_H

#include	<vector>

enum {FIND, REGISTER};
enum {FALSE, TRUE};

/* This constant depends on the representation of the # of children per node. It MUST BE a power of 2 minus 1. */
const int MAX_ARCS_PER_NODE = 255;

using namespace std;

class node;

/* Arc_node contains information describing an arc during building phase. It is then enclosed in fsa_arc structure. */
struct arc_node {
  node		*child;		/* pointer to a child node (NULL - common final node) */
  char		is_final;	/* whether labels up to that contained in this arc form a word */
  char		letter;		/* label of this arc */
};

class node;

/* Provide methods and data for handling automaton nodes. */
class node {
private:
  arc_node	*children;		/* arcs leading from node */
  int		arc_no;			/* number of the first arc of the node in all nodes of the automaton */
  int		hit_count;		/* number of arcs leading to node */
  node		*big_brother;		/* node containing all arcs of this node (and some others) */
  unsigned char	no_of_children;		/* number of those arcs */
  unsigned char	brother_offset;		/* number of the arc in big brother that is the first arc in this node */

public:
  static int	next_nodes;
  /* in write_arcs: is the node the next node in the automaton */
  int wis_next_node(const node *n) const
    { return (n != NULL && (n->big_brother ? (n->brother_offset == 0 && n->big_brother->hit_count != 0) : n->hit_count != 0)); }
  /* in number_arcs: is the node n the next node in the automaton after the
     current node; if it does not have a big brother, and it has not been
     numbered yet then we can put it right after the current node; if it has
     a big brother, and the big brother is not numbered yet, then the
     brother offset must be 0 for n to be the next node, because otherwise
     the first arc of n will not be the first arc of its big brother, so it
     will not be the next arc after the last arc of the current node. */
  int nis_next_node(const node *n) const
    { return (n != NULL && (n->big_brother ? (n->brother_offset == 0 && n->big_brother->arc_no == -1) : n->arc_no == -1)); }
  int hash(const int start, const int how_many) const;
  node(void) { children = NULL; no_of_children = 0; arc_no = -1; hit_count = 0; big_brother = NULL; brother_offset = 0; }
  node(node *n) { no_of_children = n->no_of_children; hit_count = 0;
		  arc_no = n->arc_no; big_brother = NULL; brother_offset = 0;
		  if (no_of_children) {
		    children = new arc_node[no_of_children];
		    memcpy(children, n->children, no_of_children * sizeof(arc_node));
		  }
		  else children = NULL;
		}
  ~node(void) { delete [] children; }
  arc_node *get_children(void) const { return children; }
  int get_no_of_kids(void) const { return no_of_children; }
  int get_arc_no(void) const { return arc_no; }
  void set_arc_no(const int n) { arc_no = n; }
  node *get_big_brother(void) const { return big_brother; } /*I hate C++*/
  int get_brother_offset(void) const { return brother_offset; } /* -||- */
  friend int cmp_nodes(const node *node1, const node *node2);
  friend int part_cmp_nodes(const node *small_node, const node *big_node, const int offset, const int group_size);
  friend int get_pseudo_offset(const node *small_node, const node *big_node);
  friend node * register_subnodes(node *c_node, const int group_size);
  node *add_postfix(const char *postfix);
  node *add_child(char letter, node *next_node);
  int fertile(void) const;
  int hit_node(int h = 1) { return hit_count += h; }
  node *compress_or_register(void);
  int number_arcs(const int gtl);
  int write_arcs(ostream &outfile);
  friend int check_tails(node *n, const int tail_size);
public:
  static int	no_of_arcs;	/* number of arcs generated */
};

void print_node(const node *n);
void delete_branch(node *branch);
node * mark_inner(node *n, const int no);

enum { PRIM_INDEX, SECOND_INDEX };

struct tree_index {
  union down_pointer {
    node	**leaf;		// to the table where each entry contains a substring of the given length,
                                // given number of children, and begins with the given letter.
                                // This table contains only pointers to nodes, and is sorted alphabetically
    tree_index	*indx;		// points to a table of lower level
  } down;
  int		counter;	// number of entries in *down
  int		allocated;	// number of places allocated in leaf
};

tree_index * get_index_by_name(const int index_name);
tree_index * register_at_level(const int discriminator, tree_index &table_node,
		  const int to_register, const int cluster_size = MAX_ARCS_PER_NODE);
int cmp_nodes(const node *node1, const node *node2);
int part_cmp_nodes(const node *small_node, const node *big_node, const int offset=-1, const int group_size=-1);

/* Purpose:	Compares two nodes.
 * Parameters:	node1			- (i) first node;
 *		node2			- (i) second node.
 * Returns:	< 0 if node1 < node2,
 *		= 0 if node1 = node2,
 *		> 0 if node1 > node2.
 * Remarks:	It is a wrap-around for cmp_nodes, necessary for
 *		bsearch_reg. Two spurious parameters are for compatibility
 *		with part_cmp_nodes. */
inline int a_cmp_nodes(const node *node1, const node *node2, const int, const int) { return cmp_nodes(node1, node2); }

node ** bsearch_reg(const node *n, node **low, node **high,
          int (*cmp)(const node *, const node *, int, int), int offset, int group_size);
node * find_or_register(node *n, const int index_root_name, const int register_it);

#endif
